/*	$Id: gdevprsc.c,v 1.9 2000/07/15 15:44:17 itohy Exp $	*/

/*
 * Copyright (c) 2000 ITOH, Yasufumi.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary forms are unlimited.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Prescribe --- Printer language by Kyocera
 *
 * prescr	Kyocera Prescribe II using decimal raster format
 *		300dpi, no duplex
 * prescr2	Kyocera Prescribe II using 3char/2byte encoded raster format
 *		600dpi, duplex support
 * prescrt	Kyocera Prescribe II using tiff encoded raster format
 *		600dpi, duplex support
 */
/*
 * Supported options
 *
 * -sDEVICE=<devname>
 *	<devname>: prescr, prescr2, prescrt
 *
 * -r<dpi>
 *	<dpi>:	prescr:  300 (default), 150, 75
 *		prescr2: 600, 300 (default), 200, 150, 75
 *		prescrt: 600, 300 (default), 200, 150, 75
 *
 * -dNumCopies=<number>
 *	set number of copies
 *
 * -dCasset=<cassette>
 *	<cassette>:	-1	(default) Use printer default
 *			 0	Manual feed
 *			 1	Cassette 1 / printer's cassette
 *			 2	Cassette 2
 *			 3	Cassette 3
 *			 4	Cassette 4
 *			 5	Cassette 5
 *			 6	Cassette 6
 *			99	EF-1/UF-1 (envelope/universal feeder)
 *			-2	(compatible) Manual feed
 *			-3	(compatible) Cassette 1 / printer's cassette
 *			-4	(compatible) Cassette 2
 *			-5	(compatible) Cassette 3
 *			-6	(compatible) envelope/universal feeder
 *
 * -dManualFeed
 *	Same as -dCasset=0 (overrides -dCasset option).
 *	If one of -dCasset=-2 to -dCasset=-6 is specified this option
 *	become to be same as -dCasset=-2.
 *
 * -dDuplex
 *	Enable duplex
 *
 * -dTumble
 *	Short-edge binding if Duplex
 *
 * -sEcoPrint=<lev>
 *	<lev>:	0, off:		Ecoprint off (do not save toner)
 *		1, light:	light (max saving of toner)
 *		2, medium, on:	medium
 *		3, dark:	dark (min saving of toner)
 */

#include <string.h>
#include "gdevprn.h"

#define prn_copies_device_body_rest_(print_page_copies, dupset) \
    { 0 },				/* std_procs */			\
    { 0 },				/* skip */			\
    /* gx_printer_device_procs */					\
    { NULL,				/* prn_dev_proc_print_page */	\
	print_page_copies,		/* prn_dev_proc_print_page_copies */\
	/* buf_procs */							\
	{   gx_default_create_buf_device,				\
	    gx_default_size_buf_device,					\
	    gx_default_setup_buf_device,				\
	    gx_default_destroy_buf_device				\
	},								\
	gdev_prn_default_get_space_params,/* prn_dev_proc_get_space_params */\
	gx_default_start_render_thread,	/* prn_dev_proc_start_render_thread */\
	gx_default_open_render_device,	/* prn_dev_proc_open_render_device */\
	gx_default_close_render_device,	/* prn_dev_proc_close_render_device */\
	gx_default_buffer_page		/* prn_dev_proc_buffer_page */	\
    },									\
    /* gdev_prn_space_params */						\
    {	PRN_MAX_BITMAP,			/* MaxBitmap */			\
	PRN_BUFFER_SPACE,		/* BufferSpace */		\
	/* gx_band_params_t */						\
	{ BAND_PARAMS_INITIAL_VALUES },					\
	0,				/* params_are_read_only */	\
	BandingAuto			/* banding_type */		\
    },									\
    { 0 },		/* fname */					\
    0, 0,		/* OpenOutputFile, ReopenPerPage */		\
    0, (dupset),	/* Duplex, Duplex_set */			\
    0, 0, 0, 0,		/* file_is_new, file, buffer_space, buf */	\
    0, 0,		/* buffer_memory, bandlist_memory */		\
    0,			/* free_up_bandlist_memory */			\
    0, 0,		/* page_queue, is_async_renderer */		\
    0, 0,		/* async_renderer, clist_disable_mask */	\
    { 0 }		/* orig_procs */

#define prn_copies_device_std_margins_body(dt, pr, dn, w10, h10, xr, yr, lo, to, lm, bm, rm, tm, cb, print_page_copies, dupset) \
	std_device_std_color_full_body_type(dt, &pr, dn,		\
	    &st_device_printer,						\
	    (int)((long)(w10)*(xr) / 10), (int)((long)(h10)*(yr) / 10),	\
	    xr, yr, cb,							\
	    -(lo) * (xr), -(to) * (yr),					\
	    (lm) * 72.0, (bm) * 72.0, (rm) * 72.0, (tm) * 72.0),	\
	prn_copies_device_body_rest_(print_page_copies, dupset)

#define prescribe_device(dt, pr, dn, w10, h10, xr, yr, lm, bm, rm, tm, cb, print_page_copies, dupset) \
{ prn_copies_device_std_margins_body(dt, pr, dn, w10, h10, xr, yr,	\
		lm, tm, lm, bm, rm, tm, cb, print_page_copies, dupset),	\
    -1,		/* Casset */						\
    -1,		/* EcoPrint */						\
    0,		/* ManualFeed */					\
    0		/* Tumble */						\
}

typedef struct gx_device_prescr_s {
    gx_device_common;
    gx_prn_device_common;

    /* device specific */
    int Casset;
    int EcoPrint;	/* -1: not specified, 0-3: EcoPrint */
    bool ManualFeed;
    bool Tumble;
} gx_device_prescr;

#define PRESCR_OPTION_TUMBLE		"Tumble"
#define PRESCR_OPTION_ECOLEVEL		"EcoPrint"
#define PRESCR_OPTION_MANUALFEED	"ManualFeed"
#define PRESCR_OPTION_CASSETTE		"Casset"

private const char *const ecostr[] = {
    "",
    "off", "light", "medium", "dark"
};

#define PRESCR_DPI_DEFAULT	300
#define PRESCR2_DPI_DEFAULT	300

/* The device descriptors */
private dev_proc_open_device(prescr_open);
private dev_proc_open_device(prescr2_open);

private dev_proc_close_device(prescr_close);

private dev_proc_get_params(prescr_get_params);
private dev_proc_put_params(prescr_put_params);

private dev_proc_print_page_copies(prescr_print_page_copies);
private dev_proc_print_page_copies(prescr2_print_page_copies);
private dev_proc_print_page_copies(prescrt_print_page_copies);

private const gx_device_procs prescr_prn_procs =
    prn_params_procs(prescr_open, gdev_prn_output_page, prescr_close,
			prescr_get_params, prescr_put_params);

gx_device_prescr gs_prescr_device =
    prescribe_device(gx_device_prescr, prescr_prn_procs, "prescr",
		DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS,
		PRESCR_DPI_DEFAULT, PRESCR_DPI_DEFAULT,
		5.5 / 25.4,		/* left margin */
		5.0 / 25.4,		/* bottom margin */
		5.0 / 25.4,		/* right margin */
		10.5 / 25.4,		/* top margin */
		1, prescr_print_page_copies,
		-1);			/* no duplex */

private const gx_device_procs prescr2_prn_procs =
    prn_params_procs(prescr2_open, gdev_prn_output_page, prescr_close,
			prescr_get_params, prescr_put_params);

gx_device_prescr gs_prescr2_device =
    prescribe_device(gx_device_prescr, prescr2_prn_procs, "prescr2",
		DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS,
		PRESCR2_DPI_DEFAULT, PRESCR2_DPI_DEFAULT,
		5.5 / 25.4,		/* left margin */
		5.0 / 25.4,		/* bottom margin */
		5.0 / 25.4,		/* right margin */
		10.5 / 25.4,		/* top margin */
		1, prescr2_print_page_copies,
		0);			/* duplex support */

gx_device_prescr gs_prescrt_device =
    prescribe_device(gx_device_prescr, prescr2_prn_procs, "prescrt",
		DEFAULT_WIDTH_10THS, DEFAULT_HEIGHT_10THS,
		PRESCR2_DPI_DEFAULT, PRESCR2_DPI_DEFAULT,
		5.5 / 25.4,		/* left margin */
		5.0 / 25.4,		/* bottom margin */
		5.0 / 25.4,		/* right margin */
		10.5 / 25.4,		/* top margin */
		1, prescrt_print_page_copies,
		0);			/* duplex support */

/* Printer types */
enum prescr_type {
    PRESCR,
    PRESCR2,
    PRESCRT
};

/* local functions */
private int prescribe_open(P2(gx_device *pdev, enum prescr_type ptype));
private void prescribe_line_tiff(P4(FILE *prn_stream, const byte *uncmp,
		int len, byte *cmp));
private int prescribe_print_tiff(P4(gx_device_printer *pdev, FILE *prn_stream,
		byte *img, int bpl));
private int prescribe_print_raw(P5(gx_device_printer *pdev, FILE *prn_stream,
		byte *img, int bpl, enum prescr_type ptype));
private int prescribe_paper_select(P1(gx_device_printer *pdev));
private int prescribe_print_page_copies(P4(gx_device_printer *pdev,
		FILE *prn_stream, int numcopies, enum prescr_type ptype));

/* Open the printer */
private int
prescribe_open(gx_device *pdev, enum prescr_type ptype)
{
    int xdpi = pdev->x_pixels_per_inch;
    int ydpi = pdev->y_pixels_per_inch;

    /* Resolution Check */
    if (xdpi != ydpi)
	return_error(gs_error_rangecheck);
    else if (ptype == PRESCR) {
	if (xdpi != 300 && xdpi != 150 && xdpi != 75)
	    return_error(gs_error_rangecheck);
    } else {
	if (xdpi != 600 && xdpi != 300 && xdpi != 200
	    && xdpi != 150 && xdpi != 75)
	    return_error(gs_error_rangecheck);
    }

    return gdev_prn_open(pdev);
}

/* open prescr */
private int
prescr_open(gx_device *pdev)
{
    return prescribe_open(pdev, PRESCR);
}

/* open prescr2 or prescrt */
private int
prescr2_open(gx_device *pdev)
{
    return prescribe_open(pdev, PRESCR2);
}

private int
prescr_close(gx_device *pdev)
{
    gx_device_printer *const ppdev = (gx_device_printer *) pdev;
#if 0
    gx_device_prescr *const prescr = (gx_device_prescr *) pdev;
#endif

    gdev_prn_open_printer(pdev, 1);

    fprintf(ppdev->file,
		"!R!"			/* begin Prescribe mode */
		"RES;"			/* eject page and reset */
		"EXIT,E;\n"		/* end Prescribe mode */
		);

    return gdev_prn_close(pdev);
}

/* driver -> Ghostscript */
private int
prescr_get_params(gx_device *pdev, gs_param_list *plist)
{
    gx_device_prescr *const prescr = (gx_device_prescr *) pdev;
    gs_param_string eco;
    int code;

    if ((code = gdev_prn_get_params(pdev, plist)) < 0)
	return code;

    if ((code = param_write_bool(plist, PRESCR_OPTION_TUMBLE,
				&prescr->Tumble)) < 0)
	return code;

    if ((code = param_write_bool(plist, PRESCR_OPTION_MANUALFEED,
				&prescr->ManualFeed)) < 0)
	return code;

    if ((code = param_write_int(plist, PRESCR_OPTION_CASSETTE,
				&prescr->Casset)) < 0)
	return code;

    eco.data = ecostr[prescr->EcoPrint + 1];
    eco.size = strlen(eco.data);
    eco.persistent = false;
    if ((code = param_write_string(plist, PRESCR_OPTION_ECOLEVEL, &eco)) < 0)
	return code;

    return code;
}

/* Ghostscript -> driver */
private int
prescr_put_params(gx_device *pdev, gs_param_list *plist)
{
    gx_device_prescr *const prescr = (gx_device_prescr *) pdev;
    gs_param_string eco;
    int ecolv;
    int code;

    if ((code = param_read_bool(plist, PRESCR_OPTION_TUMBLE,
				&prescr->Tumble)) < 0) {
	param_signal_error(plist, PRESCR_OPTION_TUMBLE, code);
	return code;
    }

    if ((code = param_read_bool(plist, PRESCR_OPTION_MANUALFEED,
				&prescr->ManualFeed)) < 0) {
	param_signal_error(plist, PRESCR_OPTION_MANUALFEED, code);
	return code;
    }

    if ((code = param_read_int(plist, PRESCR_OPTION_CASSETTE,
				&prescr->Casset)) < 0) {
	param_signal_error(plist, PRESCR_OPTION_CASSETTE, code);
	return code;
    }
    if (prescr->Casset < -6 || prescr->Casset > 99) {
	code = gs_error_rangecheck;
	goto errret;
    }

    if ((code = param_read_string(plist, PRESCR_OPTION_ECOLEVEL, &eco)) == 0) {
	if (eco.size == 0)	/* empty string */
	    ecolv = -1;
	else if (eco.size == 1 && (eco.data[0] >= '0' && eco.data[0] <= '3'))
	    ecolv = eco.data[0] - '0';
	else if (eco.size == 2 && !strncmp(eco.data, "on", 2))
	    ecolv = 2;		/* medium */
	else if (eco.size == 3 && !strncmp(eco.data, "off", 3))
	    ecolv = 0;
	else if (eco.size == 4 && !strncmp(eco.data, "dark", 4))
	    ecolv = 3;
	else if (eco.size == 5 && !strncmp(eco.data, "light", 5))
	    ecolv = 1;
	else if (eco.size == 6 && !strncmp(eco.data, "medium", 6))
	    ecolv = 2;
	else {
	    code = gs_error_rangecheck;
	    goto errret;
	}
    } else if (code == 1) {
	ecolv = -1;
    } else {
    errret:
	param_signal_error(plist, PRESCR_OPTION_ECOLEVEL, code);
	return code;
    }
    prescr->EcoPrint = ecolv;

    if ((code = gdev_prn_put_params(pdev, plist)) < 0)
	return code;

    return code;
}

/*
 * output a line using TIFF compression
 */
private void
prescribe_line_tiff(FILE *prn_stream, const byte *uncmp, int len, byte *cmp)
{
    const byte *src;
    byte *ptag, *p;
    int run, last;
    enum { TIF_RAW, TIF_RUN } cstat;
#define TIF_RAW_MAX	128
#define TIF_RAWLEN_TO_TAG(l)	((l) - 1)

#define TIF_RUN_MAX	128
#define TIF_RUNLEN_TO_TAG(l)	(-(unsigned)((l) - 1))

    /*
     * Algorithm:
     *	1. init:  Use RAW by default (go to 2).
     *	2. RAW: If encounter a sequence of 3 same character, switch to RUN (3).
     *		Otherwise, continue RAW (2).
     *	3. RUN: If encounter a different character, switch to RAW (2).
     *		If reach to the max count, switch to RAW (2).
     *		Otherwise, continue RUN (3).
     */
    /*
     * That is,
     *
     * abccdef	 7abccdef	(better -- use this one)
     *		 2ab2c3def
     *
     * abcccdef	 8abcccdef
     *		 2ab3c3def	(use this one to reduce backward reference)
     *
     * abccccdef 9abccccdef
     *		 2ab4c3def	(better -- use this one)
     */
    ptag = cmp;
    p = ptag + 1;
    last = 0x100;	/* do not match any byte */
    run = 0;
    cstat = TIF_RAW;

    for (src = uncmp; len > 0; len--, src++) {
	if (cstat == TIF_RAW) {
	    /*
	     *  abcc  and *src == 'c'
	     * t   lp	t: *ptag, l: last, p: *p
	     *
	     * -> 2ab3c
	     *
	     *  cc    and *src == 'c'
	     * t lp
	     *
	     * -> 3c
	     */
	    if (*src == last && run >= 2 && p[-2] == last) {
		if (run > 2) {
		    *ptag = TIF_RAWLEN_TO_TAG(run - 2);
		    ptag = p - 2;
		}
		p = ptag + 1 + 1;
		run = 3;
		cstat = TIF_RUN;
	    } else {
		if (run == TIF_RAW_MAX) {
		    *ptag = TIF_RAWLEN_TO_TAG(TIF_RAW_MAX);
		    ptag = p;
		    p = ptag + 1;
		    run = 0;
		}
		*p++ = *src;
		run++;
	    }
	} else {
	    /*
	     * cstat == TIF_RUN
	     */
	    if (*src == last) {
		run++;
		if (run == TIF_RUN_MAX) {
		    *ptag = TIF_RUNLEN_TO_TAG(run);
		    ptag = p;
		    p = ptag + 1;
		    run = 0;
		    cstat = TIF_RAW;
		}
	    } else {
		*ptag = TIF_RUNLEN_TO_TAG(run);
		ptag = p;
		p = ptag + 1;
		*p++ = *src;
		run = 1;
		cstat = TIF_RAW;
	    }
	}

	last = *src;
    }

    if (run) {
	*ptag = (cstat == TIF_RAW) ?
		    TIF_RAWLEN_TO_TAG(run) : TIF_RUNLEN_TO_TAG(run);
	ptag = p;
    }

    fprintf(prn_stream, "%d,", ptag - cmp);
    for (p = cmp; p < ptag; p++) {
	putc(*p, prn_stream);
    }
    fprintf(prn_stream, ";\n");
}

private int
prescribe_print_tiff(gx_device_printer *pdev, FILE *prn_stream,
			byte *img, int bpl)
{
    byte *pend;
    int lcnt, blanks;
    byte *cmp;
    const int cmpsz = bpl / 4 * 5 + 10;	/* would be enough */

    if ((cmp = (byte *)gs_malloc(cmpsz, 1, "prescribe_print_tiff")) == NULL)
	return_error(gs_error_VMerror);

    fprintf(prn_stream, "RVCD 2;\n"); /* receive TIFF compressed raster data */

    for (lcnt = blanks = 0; lcnt < pdev->height; lcnt++) {
	/* get one line image */
	gdev_prn_copy_scan_lines(pdev, lcnt, img, bpl);

	/* find end of line */
	for (pend = img + bpl; pend > img && pend[-1] == 0; pend--)
	    ;

	if (pend == img) {
	    /*
	     * blank lines are counted and holded
	     * to cut off the blanks at the end of paper.
	     */
	    blanks++;
	    continue;
	}

	/* emit blanks */
	if (blanks) {
	    for ( ; blanks; blanks--)
		fprintf(prn_stream, "0,;");
	    fprintf(prn_stream, "\n");
	}
	blanks = 0;

	prescribe_line_tiff(prn_stream, img, pend - img, cmp);
    }

    gs_free(cmp, cmpsz, 1, "prescribe_print_tiff");
    return 0;
}

private int
prescribe_print_raw(gx_device_printer *pdev, FILE *prn_stream,
			byte *img, int bpl, enum prescr_type ptype)
{
    byte *p, *pend;
    int lcnt, blanks;

    fprintf(prn_stream, "RVRD;\n"); /* receive (uncompressed) raster data */

    for (lcnt = blanks = 0; lcnt < pdev->height; lcnt++) {
	/* get one line image */
	gdev_prn_copy_scan_lines(pdev, lcnt, img, bpl);

	/* find end of line */
	for (pend = img + bpl; pend > img && pend[-1] == 0; pend--)
	    ;

	if (pend == img) {
	    /*
	     * blank lines are counted and holded
	     * to cut off the blanks at the end of paper.
	     */
	    blanks++;
	    continue;
	}

	/* emit blanks */
	if (blanks) {
	    for ( ; blanks; blanks--)
		fprintf(prn_stream, "1,;");
	    fprintf(prn_stream, "\n");
	}
	blanks = 0;

	/* printer firmware bug workaround (from gdevkcr) */
	if (pend == img + 255 || pend == img + 256)
	    pend = img + 257;

	fprintf(prn_stream, "%d", (int) (pend - img));
	if (ptype == PRESCR) {
	    for (p = img; p < pend; p++) {
		fprintf(prn_stream, ",");
		if (*p)
		    fprintf(prn_stream, "%d", *p);
	    }
	} else {
	    fprintf(prn_stream, ";");
	    for (p = img; p < pend; p += 2) {
		/*
		 * Encode 2 byte into three characters.
		 * See "Defining Your Own Fill patterns" in chapter 1 of
		 * Prescribe II Programming Manual for details.
		 */
		/* 2byte => 6bit 6bit 4bit */
		int x = *p >> 2;
		int y = (*p & 0x03) << 4 | p[1] >> 4;
		int z = p[1] & 0x0f;

		/*
		 * If x is zero, it can be omitted.
		 * If both x and y are zeros, they can be omitted.
		 */
		if (x)
		    putc(x == 63 ? '/' : x + 64, prn_stream);
		if (x || y)
		    putc(y == 63 ? '/' : y + 64, prn_stream);
		putc(z + 48, prn_stream);
	    }
	}
	fprintf(prn_stream, ";\n");
    }
    return 0;
}

private int
prescribe_paper_select(gx_device_printer *pdev)
{
    static const struct prescrpsz {
	int code;
	int width, height;
    } psztbl[] = {
	{  1,  279,  540 },	/* Monarch (3-7/8 x 7-1/2 inches) */
	{  2,  297,  684 },	/* Business (4-1/8 x 9-1/2 inches) */
	{  3,  312,  624 },	/* International DL (11 x 22 cm) */
	{  4,  459,  649 },	/* International C5 (16.2 x 22.9 cm) */
	{  5,  522,  756 },	/* Executive (7-1/4 x 10-1/2 inches) */
	{  6,  612,  792 },	/* US Letter (8-1/2 x 11 inches) */
	{  7,  612, 1008 },	/* US Legal (8-1/2 x 14 inches) */
	{  8,  595,  842 },	/* A4 (21.0 x 29.7 cm) */
	{  9,  516,  729 },	/* JIS B5 (18.2 x 25.7 cm) */
	{ 10,  842, 1190 },	/* A3 (29.7 x 42 cm) */
	{ 11,  729, 1032 },	/* B4 (25.7 x 36.4 cm) */
	/* Ledger should be landscape? */
	{ 12,  792, 1224 },	/* US Ledger (11 x 17 inches) */
	{ 13,  421,  595 },	/* A5 (14.8 x 21 cm) */
	{ 14,  297,  421 },	/* A6 (10.5 x 14.8 cm) */
	{ 15,  363,  516 },	/* JIS B6 (12.8 x 18.2 cm) */
	{ 16,  279,  639 },	/* Commercial #9 (3-7/8 x 8-7/8 inches) */
	{ 17,  261,  468 },	/* Commercial #6 (3-5/8 x 6-1/2 inches) */
	{ 18,  501,  709 },	/* ISO B5 (17.6 x 25 cm) */
	{ 19,  842, 1274 },	/* Custom (11.7 x 17.7 inches) */
	{ 30,  649,  918 },	/* C4 (22.9 x 32.4 cm) */
	{ 31,  284,  419 },	/* Hagaki (10 x 14.8 cm) */
	{ 32,  419,  568 }	/* Ofuku-Hagaki (14.8 x 20 cm) */
    };
    int i;
    int w = pdev->MediaSize[0];
    int h = pdev->MediaSize[1];

    /* 1st pass --- allow error by 2pt */
    for (i = 0; i < sizeof psztbl / sizeof psztbl[0]; i++)
	if (psztbl[i].width - 2 <= w && psztbl[i].width + 2 >= w &&
	    psztbl[i].height - 2 <= h && psztbl[i].height + 2 >= h)
	    return psztbl[i].code;

    /* 2nd pass --- allow error +10pt, -20pt */
    for (i = 0; i < sizeof psztbl / sizeof psztbl[0]; i++)
	if (psztbl[i].width - 20 <= w && psztbl[i].width + 10 >= w &&
	    psztbl[i].height - 20 <= h && psztbl[i].height + 10 >= h)
	    return psztbl[i].code;

    return 0;	/* unknown size */
}

private int
prescribe_print_page_copies(gx_device_printer *pdev, FILE *prn_stream,
			int num_copies, enum prescr_type ptype)
{
    gx_device_prescr *const prescr = (gx_device_prescr *) pdev;
    int code = 0;
    int dpi = pdev->x_pixels_per_inch;
    int bpl = gdev_mem_bytes_per_scan_line(pdev);
    byte *img;
    int dup;
    int psz;

    dup = (pdev->Duplex && pdev->Duplex_set >= 0) ? prescr->Tumble ? 2 : 1 : 0;
    fprintf(prn_stream, "!R!");		/* begin Prescribe mode */
    if (pdev->PageCount == 0) {
	/*
	 * First page --- initialize the printer.
	 */
	fprintf(prn_stream, "RES;");	/* reset */

	/* set paper size if applicable */
	psz = prescribe_paper_select(pdev);
	if (psz)
	    fprintf(prn_stream,
		    "SPSZ %d;",		/* set paper size */
		    psz);

	/* select paper cassette */
	if (prescr->Casset <= -2) {
	    fprintf(prn_stream,
		    "CASS %d;",		/* select cassette */
			prescr->ManualFeed ? 0 : -2 - prescr->Casset);
	} else if (prescr->ManualFeed) {
	    fprintf(prn_stream, "PSRC 0;");
	} else if (prescr->Casset >= 0) {
	    fprintf(prn_stream,
		    "PSRC %d;",		/* select paper source */
			prescr->Casset);
	}

	fprintf(prn_stream,
#if 0	/* not required for now */
		"UNIT C;"		/* use metric unit */
#endif
		"STM 0;"		/* set top margin to 0 */
		"SLM 0;"		/* set left margin to 0 */
		);

	if (dup)
	    fprintf(prn_stream, "DUPX %d;", dup);	/* duplex mode */
	fprintf(prn_stream, "RESL %d;", 300 % dpi ? 600 : 300);
	if (prescr->EcoPrint >= 0)
	    fprintf(prn_stream, "EPL %d;", prescr->EcoPrint);
    } else {
	fprintf(prn_stream, "PAGE;");	/* start new page */
    }
    if (!dup || (pdev->PageCount & 1) == 0) {
	fprintf(prn_stream, "COPY %d;", num_copies);
    }

    if ((img = (byte *)gs_malloc(bpl + 3, 1, "prescribe_print_page")) == NULL)
	return_error(gs_error_VMerror);

    img[bpl] = img[bpl + 1] = img[bpl + 2] = 0;

    /* set dot resolution used by RVRD/RVCD */
    fprintf(prn_stream, "STR %d;", dpi);

    if (ptype == PRESCRT) {
	if ((code = prescribe_print_tiff(pdev, prn_stream, img, bpl)) < 0)
	    return code;
    } else {
	if ((code = prescribe_print_raw(pdev, prn_stream, img, bpl, ptype)) < 0)
	    return code;
    }

    fprintf(prn_stream,
		"ENDR;"			/* end raster data */
		"EXIT,E;\n"		/* end Prescribe mode */
		);

    gs_free(img, bpl + 3, 1, "prescribe_print_page");

    return 0;
}

private int
prescr_print_page_copies(gx_device_printer *pdev, FILE *prn_stream,
			int num_copies)
{
    return prescribe_print_page_copies(pdev, prn_stream, num_copies, PRESCR);
}

private int
prescr2_print_page_copies(gx_device_printer *pdev, FILE *prn_stream,
			int num_copies)
{
    return prescribe_print_page_copies(pdev, prn_stream, num_copies, PRESCR2);
}

private int
prescrt_print_page_copies(gx_device_printer *pdev, FILE *prn_stream,
			int num_copies)
{
    return prescribe_print_page_copies(pdev, prn_stream, num_copies, PRESCRT);
}
